/*-
 * These functions and definitions support the site format for 4.2
 * (format proposed by Dave Gerdes):
 *
 * easting|northing|[z|[d4|]...][#category_int] [ [@attr_text OR %flt] ... ]
 *
 * to allow multidimensions (everything preceding the last '|') and any
 * number of text or numeric attribute fields.
 *
 * Author: James Darrell McCauley <mccauley@ecn.purdue.edu>
 * 31 Jan 1994
 */

#include "gis.h"
#include "site.h"

#define QUOTE '"'
#define SPACE ' '
#define BSLASH 92

Site *G_new_site_struct (n_dim, n_s_att, n_d_att)
  int n_dim, n_s_att, n_d_att;

/* Allocate memory for a Site struct. Returns NULL on error. */
{
  int i;
  Site *s;

  if (n_dim < 2 || n_s_att < 0 || n_d_att < 0)
  {
    fprintf (stderr, "PROGRAMMER ERROR: invalid # dims or fields in %s\n",
	     "G_new_site_struct");
    exit (-1);
  }

  if ((s = (Site *) G_malloc (sizeof (Site))) == NULL)
    return (Site *) NULL;
  s->cat = 0;

  if (n_dim > 2)
  {
    if ((s->dim =
	 (double *) G_malloc ((n_dim - 2) * sizeof (double))) == NULL)
      return (Site *) NULL;
  }
  s->dim_alloc = n_dim - 2;

  if (n_d_att > 0)
  {
    if ((s->dbl_att =
	 (double *) G_malloc (n_d_att * sizeof (double))) == NULL)
      return (Site *) NULL;
  }
  s->dbl_alloc = n_d_att;

  if (n_s_att > 0)
  {
    if ((s->str_att =
	 (char **) G_malloc (n_s_att * sizeof (char *))) == NULL)
      return (Site *) NULL;
    else
      for (i = 0; i < n_s_att; ++i)
	if ((s->str_att[i] =
	  (char *) G_malloc (MAX_SITE_STRING * sizeof (char *))) == NULL)
	  return (Site *) NULL;
  }
  s->str_alloc = n_s_att;

  return s;
}

G_new_get_site (ptr, s)
  FILE *ptr;
  Site *s;

/* Reads ptr and returns 0 on success, -1 on EOF, -2 on other error */
{
  int cleanse_string ();
  char sbuf[MAX_SITE_LEN], *buf, errbuf[1024];
  char ebuf[128], nbuf[128], xbuff[MAX_SITE_STRING];
  int n = 0, d = 0, c = 0, scnt = 0, dim = 0, err;
  double junka, junkb;

  buf = sbuf;

  if ((buf = fgets (sbuf, 1024, ptr)) == (char *) NULL)
    return EOF;
  while (*buf == '#')		/* comment line in site file */
    if ((buf = fgets (sbuf, 1024, ptr)) == (char *) NULL)
      return EOF;

  if (buf[strlen (buf) - 1] == '\n')
    buf[strlen (buf) - 1] = (char) NULL;

  if ((err = sscanf (buf, "%[^|]|%[^|]|%*[^\n]", ebuf, nbuf)) < 2)
  {
    fprintf (stderr, "ERROR: ebuf %s nbuf %s\n", ebuf, nbuf);
    return -2;
  }

  if (!G_scan_northing (nbuf, &(s->north), G_projection ()) ||
      !G_scan_easting (ebuf, &(s->east), G_projection ()))
    return -2;

  while (*buf != '|' && *buf != (char) NULL)
    buf++;
  if (*(buf + 1) != (char) NULL)
    buf++;
  else
    return -2;
  while (*buf != '|' && *buf != (char) NULL)
    buf++;
  if (*(buf + 1) != (char) NULL)
    buf++;
  else
    return 0;

  while (G_index (buf, '|') != (char) NULL)
    if ((err = sscanf (buf, "%lf|", &(s->dim[dim++]))) < 1)
      return -2;
    else
    {
      while (*buf != '|' && *buf != (char) NULL)
	buf++;
      if (*buf == (char) NULL)
	return 0;
      if (*(buf + 1) != (char) NULL)
	buf++;
      else
	return -2;
    }

  while (*buf != (char) NULL)
  {
    switch (*buf)
    {
    case '#':
      if (c == 0)
      {
	if (sscanf (buf, "#%d", &(s->cat)) < 1)
	  return -2;
	else
	  n++;
      }
      do
	buf++;
      while (*buf != ' ' && *buf != (char) NULL);
      if (*buf == (char) NULL)
	return 0;
      if (*(buf + 1) == (char) NULL)
	return 0;
      else
	buf++;
      break;
    case '%':
      if (d >= s->dbl_alloc)
	return -2;
      if ((err = sscanf (buf, "%%%lf", &(s->dbl_att[d++]))) < 1)
	return -2;

      while (*buf != ' ' && *buf != (char) NULL)
	buf++;
      if (*(buf + 1) == (char) NULL)
	return 0;
      else
	buf++;
      break;
    case '@':
      if (*(buf + 1) == (char) NULL)
	return 0;
      else
	buf++;
    default:			/* defaults to string attribute */
      /* allow both prefixed and unprefixed strings */
      if (c >= s->str_alloc)
	return -2;

      if ((err = cleanse_string (buf)) > 0)
      {
	G_strncpy (s->str_att[c++], buf, err);
	buf += err;
      }

      while (*buf != ' ' && *buf != (char) NULL)
	buf++;
      if (*buf == (char) NULL || *(buf + 1) == (char) NULL)
	return 0;
      else
	buf++;
      break;
    }
  }
  return 0;
}

G_new_put_site (fptr, s, has_cat)
  FILE *fptr;
  Site s;
  int has_cat;

/* Writes a site to file open on fptr. */
{
  char ebuf[MAX_SITE_STRING], nbuf[MAX_SITE_STRING];
  char xbuf[MAX_SITE_STRING], buf[MAX_SITE_LEN], *p;
  static int format_double ();
  int fmt, i, j, k;

  fmt = G_projection ();

  G_format_northing (s.north, nbuf, fmt);
  G_format_easting (s.east, ebuf, fmt);
  sprintf (buf, "%s|%s|", ebuf, nbuf);
  for (i = 0; i < s.dim_alloc - 2; ++i)
  {
    format_double (s.dim[i], nbuf);
    sprintf (xbuf, "%s|", nbuf);
    G_strcat (buf, xbuf);
  }
  if (has_cat)
  {
    sprintf (xbuf, "#%d ", s.cat);
    G_strcat (buf, xbuf);
  }
  for (i = 0; i < s.dbl_alloc; ++i)
  {
    format_double (s.dbl_att[i], nbuf);
    sprintf (xbuf, "%%%s ", nbuf);
    G_strcat (buf, xbuf);
  }
  for (i = 0; i < s.str_alloc; ++i)
  {
    /* escape double quotes */
    j=k=0;
    if (G_index (s.str_att[i], QUOTE) != (char *) NULL)
    {
      while(s.str_att[i][j] != NULL)
      {
        if (s.str_att[i][j] ==QUOTE)
        {
          xbuf[k++] = BSLASH;
          xbuf[k++] = QUOTE;
        }
        else
          xbuf[k++] = s.str_att[i][j];
        j++;
      }
      xbuf[k] = NULL;
    }
    else
      G_strcpy (xbuf,s.str_att[i]);

    G_strcpy (s.str_att[i], xbuf);

    if (G_index (s.str_att[i], SPACE) != (char *) NULL)
      sprintf (xbuf, "@\"%s\" ", s.str_att[i]);
    else
      sprintf (xbuf, "@%s ", s.str_att[i]);
    G_strcat (buf, xbuf);

  }
  fprintf (fptr, "%s\n", buf);
}

static int format_double (value, buf)
  double value;
  char *buf;
{
  sprintf (buf, "%.8lf", value);
  G_trim_decimal (buf);
}


int cleanse_string (buf)
  char *buf;
{
  static char result[MAX_SITE_STRING];
  char *stop, *p;

  p = buf;

  /* find where this string terminates */
  if (G_index (buf, QUOTE) == (char) NULL)	/* if no QUOTEs, */
  {
    stop = G_index (buf, SPACE);/* then SPACE separates */
    return (int) (stop - buf);
  }
  else				/* otherwise string is in QUOTEs */
  {				/* but we must skip over escaped */
				/* (BSLASHed) QUOTEs */
    if (*p == QUOTE)
    {
      while (*p != (char) NULL) /* get rid of first QUOTE */
      {
            *p = *(p + 1);
             p++;
      }
      p=buf;
      stop = G_index (p + 1, QUOTE);
      while (*(stop - 1) == BSLASH)
	stop = G_index (++stop, QUOTE);
    }
  }
  /* remove backslashes between buf and stop */
  while ((p=G_index (buf, BSLASH)) != (char *)NULL && p <= stop )
  {
    if (*(p + 1) != (char)NULL && *(p + 1) == QUOTE)
    {
      while (*p != (char) NULL)
      {
	    *p = *(p + 1);
	    p++;
      }
          stop--;
    }
  }
  return (int) (stop - buf);
}
